--Copyright 2010-2011 Konstantinos Asimakis
--All rights reserved.
--The source is open but changes and redistribution are only allowed for non-commercial purposes.
--For commercial purposes you should contact me to obtain a license.

--[[
Typical world:

world={
	width=10,
	height=10,
	min_x=1,
	min_y=1,
	max_x=2,
	max_y=2,
	cells={
		{x=1,y=1},
		{x=2,y=2},
	},
	cell_techs={
		{name="MBSFN",power=15},
		{name="MBSFN",power=5},
	},
	users={
		{[0]={x=1,y=1},[1]={x=2,y=2}},
		{[0]={x=2,y=2},[1]={x=1,y=1}},
	}
}

--]]

do
	local TotalPower={}
	
	--local MAX_DIST_THAT_MATTERS=20000
	local MAX_POWER=20
	local POWER_STEP=0.1
	local NOISE=0.00005 --ITAN MIA TAKSI MIKROTERO
	local cyclic_prefix_length=16.67/1000/1000 --NA THIMITHO NA TO GRAPSO KAI STO PAPER
	local useful_signal_frame_length=66.7/1000/1000
	
	TotalPower.techs={"MBSFN"}
	
	local function COST_Hata_path_loss(frequency,base_antenna_height,link_distance,mobile_antenna_height) --For the boutso.
		frequency=frequency/1000/1000
		link_distance=link_distance/1000
		local a=(1.1*math.log10(frequency)-0.7)*mobile_antenna_height-(1.56*math.log10(frequency)-0.8)
		local C=0
		return -(46.3+33.9*math.log10(frequency)-13.82*math.log10(base_antenna_height)-a+(44.9-6.55*math.log10(base_antenna_height))*math.log10(link_distance)+C)
	end
	
	local function COST_231_path_loss(frequency,base_antenna_height,link_distance,mobile_antenna_height) 
		frequency=frequency/1000/1000
		link_distance=link_distance/1000
		return -(42.6+26*math.log10(link_distance)+20*math.log10(frequency))
	end
	
	local function syncronicity(primary_signal_arrival_time,secondary_signal_arrival_time)
		local t=secondary_signal_arrival_time-primary_signal_arrival_time
		if t<-useful_signal_frame_length then return 0
		elseif t<0 then return 1+t/useful_signal_frame_length
		elseif t<cyclic_prefix_length then return 1
		elseif t<cyclic_prefix_length+useful_signal_frame_length then return 1-(t-cyclic_prefix_length)/useful_signal_frame_length
		else return 0 end
	end
	
	function count_visible_users(topo_users,timestamp)
		local count=0
		for _,user in pairs(topo_users) do
			if user[timestamp] then count=count+1 end
		end
		return count
	end
	
	function TotalPower.evaluate(topology,timestamp)
		--local power_at_user_cache={}
		for cell_tech_index,cell_tech in pairs(topology.cell_techs) do
			topology.cell_user_weight[cell_tech_index]={}
			topology.cell_techs[cell_tech_index].user_indexes={}
			topology.cell_techs[cell_tech_index].user_indexes_for_mutation={}
		end
		local indexes_of_cells_transmitting={}
		local primary_cell_data_for_user_index={}
		local score,text=0,""
		local total_transmitting_power,acceptable_service_users=0,0
		for user_index,user in pairs(topology.users) do
			if user[timestamp] then
				local primary_cell_index,primary_cell_power_at_user=nil,-math.huge
				user.path_loss_to_cell={}
				for cell_index,cell in pairs(topology.cells) do
				--for neighbour_index,neighbour in pairs(user.neighbours) do
					local distance=math.sqrt((cell.x-user[timestamp].x)^2+(cell.y-user[timestamp].y)^2)
					local path_loss=COST_231_path_loss(2000000,30,distance,1.7)
					user.path_loss_to_cell[cell_index]=path_loss
					local neighbour_power_at_user=topology.cell_techs[cell_index].power*db_to_ratio(path_loss) --EDO PREPEI NA BRO TO DEFINITION TOY DB.
					--my_print(neighbour_power_at_user)
					--power_at_user_cache[neighbouring_cell_index.."|"..path_loss]=neighbour_power_at_user --EDO PREPEI NA BRO KALITERO TROPO NA HASHARO DIO NOYMERA XORIS NA XANO AKRIBIA.
					if neighbour_power_at_user>primary_cell_power_at_user then
						primary_cell_power_at_user=neighbour_power_at_user
						primary_cell_index=cell_index
					end
					topology.cell_user_weight[cell_index][user_index]=1/path_loss --EINAI OK? KSEROYME TO PATH LOSS?
				end
				primary_cell_data_for_user_index[user_index]={index=primary_cell_index,power_at_user=primary_cell_power_at_user,tech_name=topology.cell_techs[primary_cell_index].name}
				table.insert(topology.cell_techs[primary_cell_index].user_indexes,user_index)
				user.signal={}
				for cell_index=1,#topology.cell_techs do
					table.insert(user.signal,{weight=topology.cell_user_weight[cell_index][user_index],cell_index=cell_index})
				end
				table.sort(user.signal,function (a,b) return a.weight<b.weight end)
				for i=1,3 do
					table.insert(topology.cell_techs[user.signal[i].cell_index].user_indexes_for_mutation,user_index)
				end
				--indexes_of_cells_transmitting[primary_cell_index]=true
			end
		end
		for cell_tech_index,cell_tech in pairs(topology.cell_techs) do
			if cell_tech.name=="MBSFN" then indexes_of_cells_transmitting[cell_tech_index]=true end
			--indexes_of_cells_transmitting[cell_tech_index]=true
		end
		local SE_sum=0
		for user_index,user in pairs(topology.users) do
			if user[timestamp] then
				local primary_cell_data=primary_cell_data_for_user_index[user_index]
				local user_signal,user_interference_plus_noise=primary_cell_data.power_at_user,NOISE --NA SIGOYRERPSO OTI O THORIBOS MPOREI NA THEORITHEI MIDAMINOS SE SXESI ME TO INTERFERENCE.
				local primary_signal_arrival_time=math.sqrt((topology.cells[primary_cell_data.index].x-user[timestamp].x)^2+(topology.cells[primary_cell_data.index].y-user[timestamp].y)^2)/c
				--for neighbour_index,neighbour in pairs(user.neighbours) do
				for cell_index,cell in pairs(topology.cells) do
					if cell_index~=primary_cell_data_for_user_index[user_index].index then
						local secondary_signal_arrival_time=math.sqrt((cell.x-user[timestamp].x)^2+(cell.y-user[timestamp].y)^2)/c
						local constructive_weight
						if primary_cell_data.tech_name=="MBSFN" then
							if topology.cell_techs[cell_index].name=="MBSFN" then
								constructive_weight=syncronicity(primary_signal_arrival_time,secondary_signal_arrival_time)
							else
								constructive_weight=0
							end
						else
							constructive_weight=0
						end
						local power_at_user_from_this_neighbour=topology.cell_techs[cell_index].power*db_to_ratio(user.path_loss_to_cell[cell_index])
						user_signal=user_signal+constructive_weight*power_at_user_from_this_neighbour
						user_interference_plus_noise=user_interference_plus_noise+(1-constructive_weight)*power_at_user_from_this_neighbour
					end
					--my_print(user_signal.." "..user_interference_plus_noise)
				end
				--local user_SINR=math.log10(user_signal/user_interference_plus_noise) --NA SIGOYREPSO OTI AYTOS EINAI O TIPOS.
				local user_SE=0.6*math.log(1+user_signal/user_interference_plus_noise)/math.log(2)
				if user_SE>1 then acceptable_service_users=acceptable_service_users+1 end
				SE_sum=SE_sum+user_SE
				--local user_SE=user_signal-user_interference_plus_noise
			end	
		end
		for index_of_cell_transmitting in pairs(indexes_of_cells_transmitting) do
			total_transmitting_power=total_transmitting_power+topology.cell_techs[index_of_cell_transmitting].power
		end
		--[[
		if acceptable_service_users/#topology.users>=0.95 then
			score=1+1/total_transmitting_power
		else
			score=acceptable_service_users/#topology.users
		end
		-]]
		visible=count_visible_users(topology.users,timestamp)
		local satisfied_portion=acceptable_service_users/visible
		---[[
		if satisfied_portion<0.9 then
			score=satisfied_portion-math.min(1,1/SE_sum)
		elseif satisfied_portion<1 then 
			score=
				(
					(1+1/total_transmitting_power) * math.abs(satisfied_portion-0.9) +
					(satisfied_portion-math.min(1,1/SE_sum)) * math.abs(satisfied_portion-1)
				) / (1-0.9)
		else
			score=(1+1/total_transmitting_power)
		end
		--]]
		--score=satisfied_portion/total_transmitting_power
		text="S="..round(5,score).."\nU="..round(5,satisfied_portion).." P="..round(5,total_transmitting_power)
		topology.score=score
		topology.power=total_transmitting_power --### OUT OF MODEL ###--
		topology.acceptable=satisfied_portion --### OUT OF MODEL ###--
		topology.text=text
	end
	--[[
	function TotalPower.mutate(topology,X) --Randomly mutates the topology. Wear protective clothing and avoid looking like a topology in front of this function.
		--local new_tech=TotalPower.techs[math.random(1,#TotalPower.techs)]
		local starting_cell_index=math.random(1,#topology.cells)
		local total_neighbours_to_mutate=math.floor(math.random()^(1/X)*#topology.cells[starting_cell_index].neighbours) --Ayti i seira mporei na aferethei kai na mpei me tin mia stin for.
		topology.cell_techs[starting_cell_index].name=TotalPower.techs[math.random(1,#TotalPower.techs)]	--topology.cell_techs[starting_cell_index].power=round(0,clip((math.random()*2-1)^3*MAX_POWER/10+topology.cell_techs[starting_cell_index].power,0,MAX_POWER)/POWER_STEP)*POWER_STEP
		topology.cell_techs[starting_cell_index].power=clip(math.random(-1,2)*POWER_STEP+topology.cell_techs[starting_cell_index].power,0,MAX_POWER)
		--my_print(total_neighbours_to_mutate)
		--if starting_cell_index==1 then my_print(topology.cell_techs[starting_cell_index].power) end
		for neighbour_index=1,total_neighbours_to_mutate do
			local cell_tech=topology.cell_techs[topology.cells[starting_cell_index].neighbours[neighbour_index].neighbouring_cell_index]
			cell_tech.name=TotalPower.techs[math.random(1,#TotalPower.techs)]
			--cell_tech.power=round(0,clip((math.random()*2-1)^3*MAX_POWER/10+cell_tech.power,0,MAX_POWER)/POWER_STEP)*POWER_STEP
			cell_tech.power=clip(math.random(-1,2)*POWER_STEP+cell_tech.power,0,MAX_POWER)
		end
	end
	--]]
	--[[
	function TotalPower.mutate(topology,X) --Randomly mutates the topology. Wear protective clothing and avoid looking like a topology in front of this function.
		--local new_tech=TotalPower.techs[math.random(1,#TotalPower.techs)]
		local cell_index=math.random(1,#topology.cells)
		local total_to_mutate=math.random(1,2) --Ayti i seira mporei na aferethei kai na mpei me tin mia stin for.
		topology.cell_techs[cell_index].name=TotalPower.techs[math.random(1,#TotalPower.techs)]	--topology.cell_techs[cell_index].power=round(0,clip((math.random()*2-1)^3*MAX_POWER/10+topology.cell_techs[cell_index].power,0,MAX_POWER)/POWER_STEP)*POWER_STEP
		if (#topology.cell_techs[cell_index].user_indexes>0) then
			topology.cell_techs[cell_index].power=clip(math.random(-2,4)*POWER_STEP+topology.cell_techs[cell_index].power,0,MAX_POWER)
		else
			topology.cell_techs[cell_index].power=clip(math.random(-4,2)*POWER_STEP+topology.cell_techs[cell_index].power,0,MAX_POWER)
		end
		for neighbour_index=1,total_to_mutate do
			local cell_index=math.random(1,#topology.cells)
			local cell_tech=topology.cell_techs[cell_index]
			cell_tech.name=TotalPower.techs[math.random(1,#TotalPower.techs)]
			--cell_tech.power=round(0,clip((math.random()*2-1)^3*MAX_POWER/10+cell_tech.power,0,MAX_POWER)/POWER_STEP)*POWER_STEP
			if (#topology.cell_techs[cell_index].user_indexes>0) then
				cell_tech.power=clip(math.random(-0,4)*POWER_STEP+topology.cell_techs[cell_index].power,0,MAX_POWER)
			else
				cell_tech.power=clip(math.random(-4,0)*POWER_STEP+topology.cell_techs[cell_index].power,0,MAX_POWER)
			end
		end
	end
	--]]
	function TotalPower.mutate(topology,X) --Randomly mutates the topology. Wear protective clothing and avoid looking like a topology in front of this function.
		--local new_tech=TotalPower.techs[math.random(1,#TotalPower.techs)]
		for cell_index=1,#topology.cells do
			topology.cell_techs[cell_index].name=TotalPower.techs[math.random(1,#TotalPower.techs)]	
			if (#topology.cell_techs[cell_index].user_indexes_for_mutation>0) then
				topology.cell_techs[cell_index].power=clip(math.random(-2,4)*POWER_STEP+topology.cell_techs[cell_index].power,0,MAX_POWER)
			else
				topology.cell_techs[cell_index].power=clip(math.random(-4,2)*POWER_STEP+topology.cell_techs[cell_index].power,0,MAX_POWER)
			end
		end
	end
	
	function TotalPower.get_color(tech)
		if tech.power==0 then
			return {R=0,G=192,B=255}
		else
			return {G=(tech.power/MAX_POWER)^(1/1)*255,R=(tech.power/MAX_POWER)^(1/8)*255,B=(tech.power==0 and 255 or 0)}
			--return {R=math.min(255,((tech.power*2)/MAX_POWER)^(1/4)*255),G=math.max(0,(((tech.power-MAX_POWER/2)*2)/MAX_POWER)^(1/4)*255),B=(tech.power==0 and 255 or 0)}
		end
	end
	--MBSFN={R=255,G=0,B=0},PTM={R=0,G=255,B=0}
	
	function TotalPower.get_tech_text(tech)
		return "Cell: "..tech.name.." ("..round(4,tech.power).."W)"
	end
	
	function TotalPower.sanitize(topology) --Sanitizes tech data for this metric.
		--for cell_tech_index,cell_tech in pairs(topology.cell_techs) do
		if not topology.cell_techs then topology.cell_techs={} end
		if not topology.cell_user_weight then topology.cell_user_weight={} end
		for cell_index,cell in pairs(topology.cells) do
			--my_print(cell_index)
			--my_print(table_to_string(topology))
			--my_print(tostring(topology.cell_techs[cell_index]))
			--my_print(tostring(topology.cell_techs[cell_index].name))
			if not topology.cell_techs[cell_index] then topology.cell_techs[cell_index]={} end
			if not topology.cell_techs[cell_index].user_indexes then topology.cell_techs[cell_index].user_indexes={} end
			if not topology.cell_user_weight[cell_index] then topology.cell_user_weight[cell_index]={} end
			if not is_in_table(topology.cell_techs[cell_index].name,TotalPower.techs) then topology.cell_techs[cell_index].name=TotalPower.techs[math.random(1,#TotalPower.techs)] end
			if not topology.cell_techs[cell_index].power then topology.cell_techs[cell_index].power=math.random(0,MAX_POWER/POWER_STEP)*POWER_STEP end
			cell.neighbours={}
			for cell2_index,cell2 in pairs(topology.cells) do
				if cell_index~=cell2_index then
					local distance=math.sqrt((cell.x-cell2.x)^2+(cell.y-cell2.y)^2)
					table.insert(cell.neighbours,{neighbouring_cell_index=cell2_index,distance=distance})
				end
			end
			table.sort(cell.neighbours,function (a,b) return a.distance<b.distance end)
		end
	end
	
	metrics.TotalPower=TotalPower
end