--Real Position Graphical Coverage Optimizer by Tritonio
--Copyright 2010-2011 Konstantinos Asimakis
--All rights reserved.
--The source is open but changes and redistribution are only allowed for non-commercial purposes.
--For commercial purposes you should contact me to obtain a license.
--
--Web:   http://www.inshame.com
--Email: inshame at gmail dot com
--       asimakis at ceid dot upatras dot gr
--
--Note to muggles: d = a and b or c
--in 99% of the cases does this: if a then d=b else d=c end
--Also this: d = a or b
--does this: if a then d=a else d=b end
--
--So distances are in meters.

-- NOTES FOR ME:
-- na thimatai se kathe metalaksi tin mexri tora diadromi kai na min metalasei pali toys idioys akironontas praktika tin metalaksi toys.
-- na apothikeyei kata to loading to plegmatos to poies kipseles einai geitonikes kathos kai tis metaksi toys apostaseis... To mexri poses kipseles tha thimatai gia kate kipseli tha orizetai apo to max_dist_that_matters. Diladi oses theoroynte geitonikes kai kpaoion aythereto tropo tha apothiketotal_yai metaksi toys mazi me tin apostasi poy tis xorizei.

--Here be constants!

--GUI constants
mouse_operation="user"
screen_width,screen_height=500,500 --The screen will be this size.
border=20; --Border width in pixels.

--Debugging constants
--profiling_enabled=true --If you ever enable the profiler, it stops profiling and saves it's output if you right click. NO IT DOESN'T.
print_file=io.open("stdout.txt","w")

--Optimization constans
X_factor=0.1 --Changes the aggressiveness of the mutation function.
population=50 --Population for the genetic algorithm.
protected_population=10 --This many of the top rated topologies will stay unmutated in the next generation.
discarded_population=0 --This many of the bottom rated topologies will get discarded before creating a new generation.
probability_to_crossbreed=50.0
showing=1 --Subject to show initially.
halt=true --Start stopped.
metrics={}
for line in io.popen("dir /b metrics"):lines() do
	if string.find(line,"%.lua$") then dofile("metrics/"..line) end
end
optimize=next(metrics)

--Natural constants
c=299792458

--No more constants from now on. It's an order.

if profiling_enabled then
	dofile("profiler.lua")
	profiler=newProfiler()
	profiler:start()
end

function my_print(text)
	print_file:write(text.."\n")
	print_file:flush()
end

function is_in_table(element,t)
	for i,v in pairs(t) do
		if element==v then return true end
	end
	return false
end

function db_to_ratio(db)
	return 10^(db/10)
end

function hash(something) --H PAIRS EDO EINAI OK? NA TO DOKIMASO MIPOS THELEI IPAIRS. EXO LOGO!
	local ret=0
	if type(something)=="table" then
		for i,v in pairs(something) do
			ret=(ret+(1+hash(v))*(1+hash(i)))%123456789123
		end
	elseif type(something)=="number" then
		ret=(ret+something+2)%123456789123
	elseif type(something)=="string" then
		for i,v in pairs{string.byte(something,1,-1)} do
			ret=(ret+((3+v)*i))%123456789123
		end
	end
	return ret
end
	
function create_topology(users_filename,cells_filename) --Reads two files (user positions and MBSFN cells positions and creates a topology big enough to contain them)
	local new_topology={}
	local min_x,min_y,max_x,max_y=math.huge,math.huge,-math.huge,-math.huge
	for line in io.lines(users_filename) do
		if line~="" then
			local x,y,size,total,distribution=string.match(line,"([%d%.%-]+)%s+([%d%.%-]+)%s*([%d%.%-]*)%s*([%d%-]*)%s*(%w*)")
			x,y,size,total=tonumber(x),tonumber(y),tonumber(size) or 0,tonumber(total) or 1,distribution or "RADIAL"
			if x-size<min_x then min_x=x-size end
			if x+size>max_x then max_x=x+size end
			if y-size<min_y then min_y=y-size end
			if y+size>max_y then max_y=y+size end
		end
	end
	for line in io.lines(cells_filename) do
		if line~="" then
			local x,y,resolution_x,resolution_y,total_x,total_y=string.match(line,"([%d%.%-eE]+)%s+([%d%.%-eE]+)%s*([%d%.%-eE]*)%s*([%d%.%-eE]*)%s*([%d%-eE]*)%s*([%d%-eE]*)")
			x,y,resolution_x,resolution_y,total_x,total_y=tonumber(x),tonumber(y),tonumber(resolution_x) or 1,tonumber(resolution_y) or 1,tonumber(total_x) or 1,tonumber(total_y) or 1
			if x<min_x then min_x=x end
			if x+resolution_x*(total_x-1+0.5)>max_x then max_x=x+resolution_x*(total_x-1+0.5) end
			if y<min_y then min_y=y end
			if y+resolution_y*(total_y-1+0.5)>max_y then max_y=y+resolution_y*(total_y-1+0.5) end
		end
	end
	new_topology.width,new_topology.height=max_x-min_x,max_y-min_y
	new_topology.min_x,new_topology.max_x=min_x,max_x
	new_topology.min_y,new_topology.max_y=min_y,max_y
	new_topology.users={}
	for line in io.lines(users_filename) do
		local x,y,radius,total,distribution=string.match(line,"([%d%.%-eE]+)%s+([%d%.%-eE]+)%s*([%d%.%-eE]*)%s*([%d%-eE]*)%s*(%w*)")
		x,y,radius,total=tonumber(x),tonumber(y),tonumber(radius) or 0,tonumber(total) or 1,distribution or "RADIAL"
		for cur_user=1,total do
			if radius==0 then
				table.insert(new_topology.users,{x=x,y=y})
			else
				if distribution=="RADIAL" then
					local new_x,new_y,dist,angle
					dist=math.random()*radius
					angle=math.random()*math.pi*2
					new_x=math.cos(angle)*dist+x
					new_y=math.sin(angle)*dist+y
					table.insert(new_topology.users,{x=new_x,y=new_y})
				elseif distribution=="UNIFORM" then
					local new_x,new_y
					repeat
						new_x=math.random()*2-1
						new_y=math.random()*2-1
					until new_x^2+new_y^2<1
					new_x=x+radius*new_x
					new_y=y+radius*new_y
					table.insert(new_topology.users,{x=new_x,y=new_y})
				end
			end
		end
	end
	new_topology.cells={}
	new_topology.cell_techs={}
	for line in io.lines(cells_filename) do
		local x,y,resolution_x,resolution_y,total_x,total_y=string.match(line,"([%d%.%-]+)%s+([%d%.%-]+)%s*([%d%.%-]*)%s*([%d%.%-]*)%s*([%d%-]*)%s*([%d%-]*)")
		x,y,resolution_x,resolution_y,total_x,total_y=tonumber(x),tonumber(y),tonumber(resolution_x) or 1,tonumber(resolution_y) or 1,tonumber(total_x) or 1,tonumber(total_y) or 1
		for multi_x=0,total_x-1 do
			for multi_y=0,total_y-1 do
				table.insert(new_topology.cells,{x=x+multi_x*resolution_x+(multi_y%2==0 and resolution_x/2 or 0),y=y+multi_y*resolution_y})
				table.insert(new_topology.cell_techs,{})
			end
		end
	end
	return new_topology
end

function table_to_string(t,depth,used)
	used=used or {}
	local output={}
	depth=depth or 0
	local tabs=string.rep("\t",depth)
	table.insert(output,tabs.."{ ("..tostring(t)..")\n")
	if used[t] then
		table.insert(output,tabs.."\tAlready printed\n")
	else
		used[t]=true
		for i,v in pairs(t) do
			table.insert(output,tabs.."\t[")
			if type(i)=="table" then
				table.insert(output,"\n"..table_to_string(i,depth+2,used).."\n")
				table.insert(output,tabs.."\t]")
			elseif type(i)=="string" then
				table.insert(output,'"'..tostring(i)..'"]')
			elseif type(i)=="function" then
				table.insert(output,'"'..tostring(i)..(used[v] and " Already used " or "")..'"]')
				used[i]=true
			else
				table.insert(output,tostring(i).."]")
			end
			table.insert(output," = ")
			if type(v)=="table" then
				table.insert(output,"\n"..table_to_string(v,depth+2,used))
			elseif type(v)=="string" then
				table.insert(output,'"'..tostring(v)..'"')
			elseif type(v)=="function" then
				table.insert(output,tostring(v)..(used[v] and " Already used" or ""))
				used[v]=true
			else
				table.insert(output,tostring(v))
			end
			table.insert(output,",\n")
		end
	end
	table.insert(output,tabs.."}")
	return table.concat(output)
end

function round(digits,...) --It rips the space-time continuum. (Especially when run in LuaJIT)
	for i=1,arg.n do
		arg[i]=math.floor(arg[i]*10^digits+0.5)/10^digits
	end
	return unpack(arg)
end

function screen_to_grid(x,y) --Converts screen coordinates to real coordinates.
	return (x-border)*subjects[1].width/(screen_width-2*border)+subjects[1].min_x,(y-border)*subjects[1].height/(screen_height-2*border-100)+subjects[1].min_y
end

function grid_to_screen(x,y) --Converts real coordinates to screen coordinates.
	return (x-subjects[1].min_x)/subjects[1].width*(screen_width-border*2)+border,(y-subjects[1].min_y)/subjects[1].height*(screen_height-border*2-100)+border
end

function copy_topology(topology,dont_copy_techs) --Copies a topology, replicating the cell_techs part while just "linking" to the users and cells parts which are huge and constant.
	local new_topology={}
	for i,v in pairs(topology) do
		if i=="cell_techs" then 
			if dont_copy_techs then
				new_topology[i]={}
			else
				new_topology[i]=copy_table(v)
			end
		elseif i=="score" or i=="text" then
			
		else
			new_topology[i]=v
		end
	end
	return new_topology
end

function copy_table(t) --Creates table replicas. Will only work with non table indexes. No need to worry though, only lua-psychos use tables as index values.
	local new={}
	for i,v in pairs(t) do
		new[i]=(type(v)=="table" and copy_table(v) or v)
	end
	return new
end

function clip(n,low,high) --Clips a number between two others
	return math.min(math.max(n,low),high)
end

function love.mousepressed(mouse_x,mouse_y,button)
	local true_x,true_y=screen_to_grid(mouse_x,mouse_y)
	if mouse_operation=="user" then
		if button=="l" then
			local used_tables={}
			for subject_index=1,population do
				if not used_tables[subjects[subject_index].users] then
					table.insert(subjects[subject_index].users,{x=true_x,y=true_y})
					--generate_neighbour_connections(subjects[subject_index])
					used_tables[subjects[subject_index].users]=true
				end
				metrics[optimize].sanitize(subjects[subject_index])
				metrics[optimize].evaluate(subjects[subject_index])
			end
		end
		if button=="r" then
			local used_tables={}
			local closest_distance,winner=math.huge
			for user_index,user in pairs(subjects[1].users) do
				local distance=(user.x-true_x)^2+(user.y-true_y)^2 --Squared distance to be exact...
				if distance<closest_distance then
					closest_distance=distance
					winner=user_index
				end
			end
			for subject_index,subject in pairs(subjects) do
				if not used_tables[subject.users] then
					table.remove(subject.users,winner)
					used_tables[subject.users]=true
				end
				--generate_neighbour_connections(subjects[subject_index])
				metrics[optimize].sanitize(subjects[subject_index])
				metrics[optimize].evaluate(subjects[subject_index])
			end
		end
		if button=="m" then
			mouse_operation="cell"
		end
	elseif mouse_operation=="cell" then
		if button=="l" then
			local used_tables={}
			for subject_index=1,population do
				if not used_tables[subjects[subject_index].cells] then
					table.insert(subjects[subject_index].cells,{x=true_x,y=true_y})
					--generate_neighbour_connections(subjects[subject_index])
					used_tables[subjects[subject_index].cells]=true
				end
				metrics[optimize].sanitize(subjects[subject_index])
				metrics[optimize].evaluate(subjects[subject_index])
			end
		end
		if button=="r" then
			local used_tables={}
			for subject_index,subject in pairs(subjects) do
				local closest_distance,winner=math.huge
				for cell_index,cell in pairs(subject.cells) do
					local distance=(cell.x-true_x)^2+(cell.y-true_y)^2 --Squared distance to be exact...
					if distance<closest_distance then
						closest_distance=distance
						winner=cell_index
					end
				end
				if not used_tables[subject.cells] then
					table.remove(subject.cells,winner)
					used_tables[subject.cells]=true
				end
				table.remove(subject.cell_techs,winner)
				--generate_neighbour_connections(subjects[subject_index])
				metrics[optimize].sanitize(subject)
				metrics[optimize].evaluate(subject)
			end
		end
		if button=="m" then
			mouse_operation="optimizer"
		end
	elseif mouse_operation=="optimizer" then
		if button=="l" then
			optimize=next(metrics,optimize) or next(metrics)
		end
		if button=="wu" then
			showing=showing%population+1
		end
		if button=="wd" then
			showing=(showing+population-2)%population+1
		end
		if button=="r" then
			halt=(not halt) --Halts the optimization.
		end
		if button=="m" then
			mouse_operation="user"
		end
	end
end

function love.keypressed(key,unicode)
	--[[
	if key=="=" then
		local true_x,true_y=screen_to_grid(love.mouse.getPosition())
		local closest_distance,winner=math.huge
		for cell_index,cell in pairs(subjects[showing].cells) do
			local distance=(cell.x-true_x)^2+(cell.y-true_y)^2 --Squared distance to be exact...
			if distance<closest_distance then
				closest_distance=distance
				winner=cell_index
			end
		end
		lastpower,subjects[showing].cell_techs[winner].power=subjects[showing].cell_techs[winner].power,lastpower
		metrics[optimize].sanitize(subjects[showing])
		metrics[optimize].evaluate(subjects[showing])
	end
	if key=="-" then
		local true_x,true_y=screen_to_grid(love.mouse.getPosition())
		local closest_distance,winner=math.huge
		for cell_index,cell in pairs(subjects[showing].cells) do
			local distance=(cell.x-true_x)^2+(cell.y-true_y)^2 --Squared distance to be exact...
			if distance<closest_distance then
				closest_distance=distance
				winner=cell_index
			end
		end
		lastpower=subjects[showing].cell_techs[winner].power
		subjects[showing].cell_techs[winner].power=0.0005
		metrics[optimize].sanitize(subjects[showing])
		metrics[optimize].evaluate(subjects[showing])
	end
	--]]
	if key=="o" then
		overlay=not overlay
	end
	if key=="d" then
		my_print("\n\n\n-------------------\n\n\n")
		my_print(table_to_string(_G))
	end
	if key=="c" then
		collectgarbage()
	end
	if key=="p" then
		if profiling_enabled then
			profiler:stop()
			local outfile = io.open( "profile.txt", "w+" )
			profiler:report( outfile )
			outfile:close()
		end
	end
	if key=="w" then
		h=io.open("users_output.txt","w")
		for user_index,user in ipairs(subjects[showing].users) do
			h:write(user.x.." "..user.y.."\n")
		end
		h:close()
		h=io.open("cells_output.txt","w")
		for cell_index,cell in ipairs(subjects[showing].cells) do
			h:write(cell.x.." "..cell.y.."\n")
		end
		h:close()
	end
end

function love.load()
	math.randomseed(os.time())
	local map_x,map_y,meters_per_pixel=string.match(io.open("map.txt"):read(),"([%d%.%-]+)%s+([%d%.%-]+)%s+([%d%.%-]+)")
	map={x=tonumber(map_x),y=tonumber(map_y),meters_per_pixel=tonumber(meters_per_pixel),image=love.graphics.newImage("map.bmp")}
	subjects={}
	subjects[1]=create_topology("user_positions.txt","cell_positions.txt")
	metrics[optimize].sanitize(subjects[1])
	metrics[optimize].evaluate(subjects[1])
	for subject_index=2,population do
		subjects[subject_index]=copy_topology(subjects[1],true)
		metrics[optimize].sanitize(subjects[subject_index])
		metrics[optimize].evaluate(subjects[subject_index])
	end
	table.sort(subjects,
		function (one,another)
			return one.score>another.score --KLASIKOS KAGKOYRAS KAI KARAMOYZA.
		end
	)
	tries=0
	love.graphics.setMode( screen_width, screen_height, false, true, 0 )
	love.graphics.setBackgroundColor(255,255,255)
	love.graphics.setColor(0,0,0)
	--my_print(table_to_string(subjects))
end

function love.draw()
	love.graphics.setCaption("AccuGraCovO by Tritonio ("..math.floor(collectgarbage("count")/1024+0.5).."MB)")
	map_on_screen_x,map_on_screen_y=grid_to_screen(map.x,map.y)
	--my_print((subjects[1].width/screen_width) .." ".. map.meters_per_pixel)
	love.graphics.setColor(255,255,255,255)
	love.graphics.draw(map.image,map_on_screen_x,map_on_screen_y,0, (screen_width-2*border) * map.meters_per_pixel / subjects[1].width, (screen_height-2*border-100) * map.meters_per_pixel / subjects[1].height)
	for cell_index,cell in pairs(subjects[showing].cells) do
		for _,user_index in pairs(subjects[showing].cell_techs[cell_index].user_indexes) do
			local user=subjects[showing].users[user_index]
			local color={R=200,G=200,B=200,total=1}
			local screen_user_x,screen_user_y=grid_to_screen(user.x,user.y)
			local screen_cell_x,screen_cell_y=grid_to_screen(cell.x,cell.y)
			love.graphics.setColor(color.R/color.total,color.G/color.total,color.B/color.total)
			love.graphics.line(screen_user_x,screen_user_y,screen_cell_x,screen_cell_y)
		end
		local color={R=0,G=0,B=0,total=0}
		for subject_index,subject in pairs(subjects) do
			if (overlay and subject_index<=showing) or (not overlay and subject_index==showing) then
				local tech_color=metrics[optimize].get_color(subject.cell_techs[cell_index])
				color.R=color.R+tech_color.R
				color.G=color.G+tech_color.G
				color.B=color.B+tech_color.B
				color.total=color.total+1
			end
		end
		local screen_cell_x,screen_cell_y=grid_to_screen(cell.x,cell.y)
		local vertices={
			screen_cell_x-1,	screen_cell_y-6;
			screen_cell_x+1,	screen_cell_y-6;
			screen_cell_x+6,	screen_cell_y-12;
			screen_cell_x-6,	screen_cell_y-12;
			
			screen_cell_x-1,	screen_cell_y-6;
		}
		love.graphics.setColor(color.R/color.total,color.G/color.total,color.B/color.total)
		love.graphics.polygon("fill",vertices)
		local vertices={
			screen_cell_x-1,	screen_cell_y-6;
			screen_cell_x-1,	screen_cell_y;
			screen_cell_x+1,	screen_cell_y;
			screen_cell_x+1,	screen_cell_y-6;
			
			screen_cell_x-1,	screen_cell_y-6;
		}
		love.graphics.setColor(color.R/color.total,color.G/color.total,color.B/color.total)
		love.graphics.polygon("fill",vertices)
	end
	for user_index,user in pairs(subjects[showing].users) do
		local color={R=128,G=128,B=128,total=1}
		local screen_user_x,screen_user_y=grid_to_screen(user.x,user.y)
		local vertices={
			screen_user_x-2,	screen_user_y;
			screen_user_x,		screen_user_y-2;
			screen_user_x+2,	screen_user_y;
			screen_user_x,		screen_user_y+2;
			
			screen_user_x-2,	screen_user_y;
		}
		love.graphics.setColor(color.R/color.total,color.G/color.total,color.B/color.total)
		love.graphics.polygon("fill",vertices)
		--love.graphics.setColor(0,0,0)
		--love.graphics.polygon("line",vertices)
	end
	--[[
	love.graphics.print(tries.." tries.\nShowing: "..(overlay and "1-" or "")..showing.."/"..population.."\nScore: "..round(4,subjects[showing].evaluation.RE).."\nMBSFN/total: "..subjects[showing].evaluation.total_MBSFN.."/"..subjects[showing].evaluation.total_transmitting.."\nMy Ratio: "..round(6,subjects[showing].evaluation.myratio).."\n",25,screen_height-90)
	love.graphics.setColor(0,0,255)
	love.graphics.print("MBSFN",screen_width-150,screen_height-90)
	love.graphics.setColor(0,255,255)
	love.graphics.print("Assisting",screen_width-150,screen_height-75)
	love.graphics.setColor(255,0,0)
	love.graphics.print("PTM",screen_width-150,screen_height-60)
	love.graphics.setColor(0,0,0)
	--]]
	
	local true_mouse_x,true_mouse_y=screen_to_grid(love.mouse.getPosition())
	local closest_distance,winner=math.huge
	for cell_index,cell in pairs(subjects[showing].cells) do
		local distance=(cell.x-true_mouse_x)^2+(cell.y-true_mouse_y)^2 --Squared distance to be exact...
		if distance<closest_distance then
			closest_distance=distance
			winner=cell_index
		end
	end
	last_cell_text=metrics[optimize].get_tech_text(subjects[showing].cell_techs[winner])
	
	love.graphics.setColor(0,0,0)
	love.graphics.print(tries.." generations.\nShowing: "..(overlay and "1-" or "")..showing.."/"..population.."\nMouse operation: "..mouse_operation.."\nOptimizing: "..optimize.."\n"..subjects[showing].text.."\n"..(last_cell_text or ""),border,screen_height-100)
	local info_messages
	if mouse_operation=="user" then
		info_messages={
			"Left click: Place",
			"Right click: Remove",
			"Middle click: Switch mode",
			"O: Overlay subjects",
			"D: Dump subjects to file",
			"C: Collect garbage",
			"W: Write topology to file",
		}
	elseif mouse_operation=="cell" then
		info_messages={
			"Left click: Place",
			"Right click: Remove",
			"Middle click: Switch mode",
			"O: Overlay subjects",
			"C: Collect garbage",
			"W: Write topology to file",
		}
	elseif mouse_operation=="optimizer" then
		info_messages={
			"Left click: Change optimization metric",
			"Right click: Start/Stop optimization",
			"Middle click: Switch mode",
			"Wheel up: Next subject",
			"Wheel down: Previous subject",
			"O: Overlay subjects",
			"D: Dump subjects to file",
			"W: Write topology to file",
		}
	end
	local info_text
	if #info_messages>5 then
		local starting_point=(#info_messages>4 and math.floor(os.time()/3) or 0)
		info_text=
			info_messages[starting_point%#info_messages+1].."\n"..
			info_messages[(starting_point+1)%#info_messages+1].."\n"..
			info_messages[(starting_point+2)%#info_messages+1].."\n"..
			info_messages[(starting_point+3)%#info_messages+1].."\n"..
			info_messages[(starting_point+4)%#info_messages+1].."\n"
	else
		info_text=table.concat(info_messages,"\n")
	end
	love.graphics.print("Instructions:\n"..info_text,screen_width-250,screen_height-100)
end

function love.update(dt)
	if not halt then
		tries=tries+1
		local hashes={}
		for subject_index=protected_population,1,-1 do
			table.insert(hashes,hash(subjects[subject_index].cell_techs))
		end
		for subject_index=population,protected_population+1,-1 do --For the unprotected ones... (we suppose the list is sorted)
			if math.random()<probability_to_crossbreed then
				subjects[subject_index]=metrics[optimize].crossbreed(subjects[subject_index],subjects[math.random(1,math.min(subject_index,population-discarded_population))])
			else
				subjects[subject_index]=copy_topology(subjects[math.random(1,math.min(subject_index,population-discarded_population))])
			end
			metrics[optimize].mutate(subjects[subject_index],X_factor)
			metrics[optimize].evaluate(subjects[subject_index])
			local subject_hash=hash(subjects[subject_index].cell_techs)
			if is_in_table(subject_hash,hashes) then
				subjects[subject_index].cell_techs={}
				metrics[optimize].sanitize(subjects[subject_index])
				metrics[optimize].evaluate(subjects[subject_index])
			else
				table.insert(hashes,subject_hash)
			end
		end
		table.sort(subjects,
			function (one,another)
				return one.score>another.score --KLASIKOS KAGKOYRAS KAI KARAMOYZA.
			end
		)
		my_print(tries.."\t"..subjects[1].power.."\t"..subjects[1].acceptable) --### OUT OF MODEL ###--
	end
end